"""
This module defines classes for simulating network models for polymer melt rheology in viscometric flows.

References:
(1) Geurts, K.R. and L.E. Wedgewood, "A finitely extensible network strand model with nonlinear backbone
forces and entanglement kinetics," J. Chem. Phys., 1-January-1997, 106(1), pp. 339-346.
(2) Biller and Petruccione, J. Chem. Phys., 89(1), pp.577-582, 1988.  

Exported classes:
    RheoNetSim: Base class for simulators of rheology network models in viscometric flows.
    ElongateNetSim: Subclass of RheoNetSim that simulates a network strand model in elongational viscometric flow.
    RheoNetSimOut: Class to accumulate (over time steps) and then return time-averaged output from RheoNetSim.run_sim() method.
    ElongateNetSimOut: Subclass of RheoNetSimOut to accumulate (over time steps) and then return time-averaged output from RheoNetSim.run_sim() method,
                       with additional output values appropriate for elongational flow simulation.

    Note that:
    (1) A XNetSim class uses it's factory method _createSimOutObj() to create the correct type of XSimOut object.
    (2) The XSimOut classes are populated by calls to XNetSim._computeTimestepOutputs() method.

Exported functions:
    move_strand_fene_elongational: Appy Euler integration to advance a FENE strand s, one time step eps, under elongation rate gamdot.
    move_strand_fens_elongational: Appy Euler integration to advance a FENS strand s, one time step eps, under elongation rate gamdot.

Exported exceptions:
    None
"""


# Standard imports
from copy import deepcopy
from math import log, copysign
from array import array
from random import uniform
from math import sqrt
from symtable import Symbol

# Local imports
from RheologyNetworkModelSimulator.strand import Strand, FENEStrand
from RheologyNetworkModelSimulator.ensemble import Ensemble
from RheologyNetworkModelSimulator.qplot import TextPlot


class RheoNetSimOut:
    """
    This is a base class for subclasses which accumulate (over time steps) and then returns time-averaged output from RheoNetSim.run_sim().
    """
    def __init__(self):
        """
        Time-averaged output values from simulation:
            n_ave = n/no = Time-averaged, non-dimensional (relative to equilibrium number) number of entangled strands in the network ensemble, int
            piYX_ave = Time-averaged, non-dimensional YX-component of the total stress tensor, float
            Q_ave = Time-averaged, non-dimensinal ensemble-average network strand length, float
        Arrays of values over time steps in simulation output:
            n_vals = Array of strand number values in simulation output, array of int
            piYX_vals = Array of piYX values in simulation output, array of float
            Q_vals = Array of Q values in simulation output, array of float
            time_vals = Array of time values in simulation output, array of float
        """
        self.n_ave=0
        self.piYX_ave=0
        self.Q_ave=0
        self.n_vals = array('i')
        self.piYX_vals = array('f')
        self.Q_vals = array('f')
        self.time_vals = array('f')

    def __str__(self):
        """
        Convert object to str, for example, to write results to file.
        :return: Class attributes converted to a string representation, suitable for writing results to a file, as str
        """
        result = f"{self.n_vals[len(self.n_vals)-1]}, {self.piYX_vals[len(self.piYX_vals)-1]}, {self.Q_vals[len(self.Q_vals)-1]}"
        return result

    def time_step_output(self):
        """
        Generate a string representation of labeled output values for the latest time step, for example, to print to show progress of simulation.
        :return: Class attributes converted to a string representation, suitable for printing time step results to show progress of simulation, as str
        """
        result = f"n: {self.n_vals[len(self.n_vals)-1]}, piYX: {self.piYX_vals[len(self.piYX_vals)-1]}, Q: {self.Q_vals[len(self.Q_vals)-1]}"
        return result

    def final_output(self):
        """
        Generate a string representation of labeled output values for the final time-averaged results, for example, to print at the end of simulation.
        :return: Class attributes converted to a string representation, suitable for printing final results at end of simulation, as str
        """
        result = f"Avg n: {self.n_ave}\nAvg piYX: {self.piYX_ave}\nAvg Q: {self.Q_ave}\n"
        return result


class ElongateNetSimOut(RheoNetSimOut):
    """
    A class to accumulate (over time steps) and then return time-averaged output from TroutSim.run_sim().
    """
    def __init__(self):
        """
        Time-averaged output values from simulation:
            trout1_ave = Time-averaged, non-dimensional Trouton 1 viscosity, float
            trout2_ave = Time-averaged, non-dimensional Trouton 2 viscosity, float
        Arrays of values over time steps in simulation output:
            trout1_vals = Array of Trouton 1 viscosity values in simulation output, array of float
            trout2_vals = Array of Trouton 3 viscosity values in simulation output, array of float
        TextPlot objects from simulation:
            trout1_plt = TextPlot object plot of Trouton 1 viscosity
        """
        super().__init__()
        self.trout1_ave=0
        self.trout2_ave=0
        self.trout1_vals= array('f')
        self.trout2_vals= array('f')
        self.trout1_plt=None

    def __str__(self):
        """
        Convert object to str, for example, to write results to file.
        :return: Class attributes converted to a string representation, suitable for writing results to a file, as str
        """
        result = super().__str__()
        result += f", {self.trout1_vals[len(self.trout1_vals)-1]}, {self.trout2_vals[len(self.trout2_vals)-1]}"
        return result

    def time_step_output(self):
        """
        Generate a string representation of labeled output values for the latest time step, for example, to print to show progress of simulation.
        :return: Class attributes converted to a string representation, suitable for printing time step results to show progress of simulation, as str
        """
        result = super().time_step_output()
        result += f", Trout 1: {self.trout1_vals[len(self.trout1_vals)-1]}, Trout 2: {self.trout2_vals[len(self.trout2_vals)-1]}"
        return result

    def final_output(self):
        """
        Generate a string representation of labeled output values for the final time-averaged results, for example, to print at the end of simulation.
        :return: Class attributes converted to a string representation, suitable for printing final results at end of simulation, as str
        """
        result = super().final_output()
        result += f"Avg Trout 1: {self.trout1_ave}\nAvg Trout 2: {self.trout2_ave}\n\n{self.trout1_plt}"
        return result


class RheoNetSim:
    """
    Base class for simulators of rheology network models in viscometric flows.

    Uses the method of Biller and Petruccione, J. Chem. Phys., 89(1), pp.577-582, 1988.  

    Nondimensionalization Scheme:
        Strand internal coordinates are non-dimensionalized consistent with proto_strand argument to __init__().
        Loss rate is multiplied by the loss rate constant, Lambdao.
        Time step size is divided by the loss rate constant, Lamdao.
        Stress tensor is divided by NokT.
    """
    def __init__(self, si={}, proto_strand=Strand(), strand_mover=None):
        """
        Create an object used to run simulation.

        :param si: Dictionary of non-dimensional input values for the simulation as follows:
            gamdot = shear rate or elongation rate, float
            begstrand = number of strands in equilibrium ensemble, int
            eps = time step size, float
            steps = total time steps, int
            outfile = file for writing simulation steps output, string
        :param proto_strand: A prototype Strand object used to create new strands in the network ensemble, Strand subclass object
        :param strand_mover: A function that moves a Strand object according to the rheological model and viscometric flow being simulated, function
                             Function signature: strand_mover(strand, eps, gamdot, n), where
                                strand is a Strand object to move,
                                eps is the time step size, float,
                                gamdot is the shear or elongation rate, float,
        """
        assert(type(si)==dict)
        self.sim_input = si
        self._time_frac_ss = 0.25 # The fraction of simulation time the simulation is assumed to be at steady state
        assert(isinstance(proto_strand, Strand))
        self._proto_strand = proto_strand
        assert(callable(strand_mover))
        self._strand_mover = strand_mover
        self._sim_output = self._createSimOutObj()

    def _createSimOutObj(self):
        """
        Create and return an output object appropriate for this simulation class. Must be implemented by subclasses.
        Will raise NotImplementedError if called from this base class.
        :return: An output object appropriate for this simulation class, RheoNetSimOut subclass object
        """
        raise NotImplementedError("Subclasses must implement _createSimOutObj method")
        return RheoNetSimOut()

    def _getOutFileHeader(self):
        """
        Return a string to write as a header line into the output file, that is, the column headers. Should be extended as needed by subclasses.
        This would typically be done by calling super() to get the base class functionality, and then appending additional column headers by
        += f", {}" the returned string.
        :return: String for output file header, that is, the column headers, as string
        """
        result = f"Time, Strands, <Q>, piYX"
        return result

    def _computeTimestepOutputs(self, we=None, time_value=0, finalize=False):
        """
        Compute output values for a single time step. Values should be appended to the "vals" arrays of the output object.
        Should be extended as needed by subclasses. Subclasses should call super() to get base class functionality.
        :param we: Working ensemble at current time step, Ensemble object
        :param time_value: Current time value, float
        :param finalize: If True, finalize any calculations needed at the end of the simulation, bool
                         Typically, this would be used to compute time-averaged values from accumulated sums.
        :return: None. Output values are appended to the "vals" arrays of the output object.
        """
        assert(isinstance(we, Ensemble))

        if not finalize:
            # Compute and store output values for this time step, and accumulate sums for time-averaged values later

            n_val = len(we)
            self._sim_output.n_vals.append(n_val)

            stress = we.ensemble_stress()
            piYX = stress[3] / self.sim_input['begstrand']
            self._sim_output.piYX_vals.append(piYX)

            q_val = we.q_ave()
            self._sim_output.Q_vals.append(q_val)

            self._sim_output.time_vals.append(time_value)

            # If we are far enough along to assume steady state, store some information into results,
            # which will later be used to generate time-averaged values
            if time_value >= self.sim_input['eps']*self.sim_input['steps']*(1.-self._time_frac_ss):
                self._sim_output.n_ave += n_val/self.sim_input['begstrand']
                self._sim_output.piYX_ave += piYX
                self._sim_output.Q_ave += q_val

        else:
            # Finalize any calculations needed at the end of the simulation.
            # Here we will compute time-averaged values from accumulated sums, by dividing by the number of time steps used in the averaging.
            t_res = round(self.sim_input['steps']*self._time_frac_ss)
            self._sim_output.n_ave = self._sim_output.n_ave / t_res
            self._sim_output.piYX_ave = self._sim_output.piYX_ave / t_res
            self._sim_output.Q_ave = self._sim_output.Q_ave / t_res

        return None
        
    def run_sim(self):
        """
        Execute the simulation, following a template design.
        :return: Results from the simulation, as a RheoNetSimOut object
        """
        #  Open a file for writing output from steps
        out_f = open(self.sim_input['outfile'], 'w')

        # Write a header line into the files with column headers
        header_str = self._getOutFileHeader() + '\n'
        out_f.write(f"{header_str}")

        # The equilibrium ensemble is used for selecting new strands to join the network.
        # It is generated using the prototype strand provided when the simulation object was created.
        ee = Ensemble(self._proto_strand.generate_eq_ensemble(self.sim_input['begstrand']))
        # TODO: Fix so that uses any path provided by sim_input['outfil']
        ee.write_ensemble_to_file('eq_ensemble_strands.csv')

        # Make a deep copy of the equilibrium  ensemble as the initial working ensemble.
        # It is important that this is a deep copy, so that the Strand objects themselves are not shared between ee and we.
        # Otherwise, as the simulation proceeded, the Strand objects in ee would be modified (e.g., deformed by movement) as well as those in we,
        # and it would no longer represent an equilibrium ensemble.
        we=deepcopy(ee)

        # For each time step of the simulation ...
        for ts in range(1, self.sim_input['steps']):

            # Advance each strand over one time step
            for s in we:
                self._strand_mover(s, self.sim_input['eps'], self.sim_input['gamdot'])

            # Check survival of each strand.
            # Since we.strands is a copy of the list of strands in we, but the Strand objects in the list themselves are shared,
            # the call to we.remove(s) below removes the correct Strand object from we, but the iteration over s in we.strands is unaffected.
            # TODO: Try list comprehension instead? Or a list ot strands to remove afterwards.
            for s in we.strands:
                p1 = uniform(0.0, 1.0)
                if p1 <= s.loss_prob(self.sim_input['eps']):
                    we.remove(s)

            # Create new strands to enter the network, from the equilibrium ensemble
            for s in ee:
                p1 = uniform(0.0, 1.0)
                if p1 <= s.loss_prob(self.sim_input['eps']):
                    # Here we create a new Strand object, ns, to add to the working ensemble we. It starts as a copy of
                    # strand s from the equilibrium ensemble, and thus starts with the same internal coordinates as strand s.
                    # However, from then on, it deforms independently.
                    ns = deepcopy(s)
                    # Move the new strand for the portion of the part of a time step for which it existed
                    p1 = uniform(0.0, 1.0)
                    y = -log(1. - p1 * ns.loss_prob(self.sim_input['eps'])) / ns.loss_rate()
                    self._strand_mover(ns, y, self.sim_input['gamdot'])

                    we.append(ns)

            # Compute some output values for this time step
            self._computeTimestepOutputs(we, ts * self.sim_input['eps'])

            # Print some output to the stdout
            print(f"Time: {round(ts * self.sim_input['eps'], 9)}, {self._sim_output.time_step_output()}")

            # Write some output to a file
            out_f.write(f"{round(ts * self.sim_input['eps'], 9)}, {self._sim_output}\n")

        out_f.close()

        # Finalize results to return
        self._computeTimestepOutputs(we, 0, finalize=True)

        return self._sim_output


class ElongateNetSim(RheoNetSim):
    """
    Simulates a network strand model in elongational viscometric flow.

    Uses the method of Biller and Petruccione, J. Chem. Phys., 89(1), pp.577-582, 1988.  

    Nondimensionalization Scheme:
        Strand internal coordinates are non-dimensionalized consistent with proto_strand argument to __init__().
        Loss rate is multiplied by the loss rate constant, Lambdao.
        Time step size is divided by the loss rate constant, Lamdao.
        Stress tensor is divided by NokT.
    """
    def __init__(self, si={}, proto_strand=Strand(), strand_mover=None):
        """
        Create an object used to run simulation.

        :param si: Dictionary of non-dimensional input values for the simulation as follows:
            gamdot = elongation rate, float
            begstrand = number of strands in equilibrium ensemble, int
            eps = time step size, float
            steps = total time steps, int
            outfile = file for writing simulation steps output, string
        """
        super().__init__(si, proto_strand, strand_mover)

    def _createSimOutObj(self):
        """
        Create and return an output object appropriate for this simulation class.
        :return: An output object appropriate for this simulation class, RheoNetSimOut subclass object
        """
        return ElongateNetSimOut()

    def _getOutFileHeader(self):
        """
        Return a string to write as a header line into the output file, that is, the column headers.
        :return: String for output file header, that is, the column headers, as string
        """
        result = super()._getOutFileHeader()
        result += f", trouton_1, trouton_2"
        return result

    def _computeTimestepOutputs(self, we=None, time_value=0, finalize=False):
        """
        Compute output values for a single time step.
        :param we: Working ensemble at current time step, Ensemble object
        :param time_value: Current time value, float
        :param finalize: If True, finalize any calculations needed at the end of the simulation, bool
                         Typically, this would be used to compute time-averaged values from accumulated sums.
        :return: None. Output values are appended to the "vals" arrays of the output object.
        """
        assert(isinstance(we, Ensemble))
        super()._computeTimestepOutputs(we, time_value, finalize)

        if not finalize:
            # Compute and store output values for this time step, and accumulate sums for time-averaged values later

            stress = we.ensemble_stress()
            trout1 = (stress[2] - stress[0]) / self.sim_input['begstrand'] / self.sim_input['gamdot']
            self._sim_output.trout1_vals.append(trout1)
            trout2 = (stress[1] - stress[0]) / self.sim_input['begstrand'] / self.sim_input['gamdot']
            self._sim_output.trout2_vals.append(trout2)


            # If we are far enough along to assume steady state, store some information into results,
            # which will later be used to generate time-averaged values
            if time_value >= self.sim_input['eps']*self.sim_input['steps']*(1.-self._time_frac_ss):
                self._sim_output.trout1_ave+=trout1
                self._sim_output.trout2_ave+=trout2

        else:
            # Finalize any calculations needed at the end of the simulation.
            # Here we will compute time-averaged values from accumulated sums, by dividing by the number of time steps used in the averaging.
            t_res = round(self.sim_input['steps']*self._time_frac_ss)
            self._sim_output.trout1_ave = self._sim_output.trout1_ave / t_res
            self._sim_output.trout2_ave = self._sim_output.trout2_ave / t_res

            # Make a TextPlot object for plotting elongational viscosity vs time
            _symbol = array('u')
            _symbol.append('O')
            _npts=len(self._sim_output.trout1_vals)
            _x=[self._sim_output.time_vals]
            _y=[self._sim_output.trout1_vals]
            _titl1='FENE Network Start-up of Elongational Flow Simulation'
            _titl2='O: Elongational Viscosity vs Time'
            self._sim_output.trout1_plt = TextPlot(ncur=1, npts=_npts, x=_x, y=_y, symbol=_symbol, titl1=_titl1, titl2=_titl2)

        return None


# Appy Euler integration to advance a strand s, one time step eps, under elongation rate gamdot.
def move_strand_fene_elongational(s, eps, gamdot):
    """
    Apply Euler integration to advance the internal coordinates of a network strand by one time step.
    :param s: Network Strand to advance, as Strand object
    :param eps: Time step size, as float
    :param gamdot: Elongation rate, as float
    :return: None. Strand s's internal coordinates are updated upon return.
    """

    alpha = 1.0 - pow(sqrt(s.str_len_sqr()), s.n)

    # Actual movement of the strand by Euler integration
    s.qx = s.qx - alpha * gamdot * .5 * s.qx * eps
    s.qy = s.qy - alpha * gamdot * .5 * s.qy * eps
    s.qz = s.qz + alpha * gamdot * s.qz * eps

    if s.max_length is not None:
        if sqrt(s.str_len_sqr()) > s.max_length:
            s.qz = copysign(sqrt(0.9999*s.max_length - s.qy * s.qy - s.qx * s.qx), s.qz)


# Appy Euler integration to advance a strand s, one time step eps, under elongation rate gamdot.
def move_strand_fens_elongational(s, eps, gamdot):
    """
    Apply Euler integration to advance the internal coordinates of a network strand by one time step.
    :param s: Network Strand to advance, as Strand object
    :param eps: Time step size, as float
    :param gamdot: Elongation rate, as float
    :return: None. Strand s's internal coordinates are updated upon return.
    """

    alpha = 1.0 - pow(sqrt(s.str_len_sqr()/s.max_length), 2.0)

    # Actual movement of the strand by Euler integration
    s.qx = s.qx - alpha * gamdot * .5 * s.qx * eps
    s.qy = s.qy - alpha * gamdot * .5 * s.qy * eps
    s.qz = s.qz + alpha * gamdot * s.qz * eps

    if s.max_length is not None:
        if sqrt(s.str_len_sqr()) > s.max_length:
            s.qz = copysign(sqrt(0.9999*s.max_length - s.qy * s.qy - s.qx * s.qx), s.qz)



"""
This module contains unit tests for the TextPlot class and the qplot function.
"""


# standard imports
import unittest
from array import array
import sys
from io import StringIO

# local imports
from RheologyNetworkModelSimulator.qplot import TextPlot, qplot


class Test_qplot_tests(unittest.TestCase):
    def test_TextPlot_qplot(self):
        self.maxDiff = None
        ncur=2
        npts=10
        x=[array('f'),array('f')]
        y=[array('f'),array('f')]
        for i in range(npts):
            x[0].append(float(i))
            x[1].append(float(i))
            y[0].append(float(i*i))
            y[1].append(float(i*i*i))
        symbol=array('u')
        symbol.append('*')
        symbol.append('o')
        titl1='Test ASCII Plot'
        titl2='*: y = x^2; o: y = x^3'
        plot=TextPlot(ncur,npts,x,y,symbol,titl1,titl2)
        act_val = plot._qplot()
        exp_val=[]
        exp_val.append('Test ASCII Plot')
        exp_val.append('*: y = x^2; o: y = x^3')
        exp_val.append(' ')
        exp_val.append('0.00e+00.......2.25e+00.......4.50e+00.......6.75e+00.......9.00e+00')
        exp_val.append('|--------------|--------------|--------------|--------------o  729.00')
        exp_val.append('|                                                           |')
        exp_val.append('|                                                           |')
        exp_val.append('|                                                           |')
        exp_val.append('|                                                           |')
        exp_val.append('+                                                           +  546.75')
        exp_val.append('|                                                    o      |')
        exp_val.append('|                                                           |')
        exp_val.append('|                                                           |')
        exp_val.append('|                                                           |')
        exp_val.append('+                                                           +  364.50')
        exp_val.append('|                                              o            |')
        exp_val.append('|                                                           |')
        exp_val.append('|                                                           |')
        exp_val.append('|                                       o                   |')
        exp_val.append('+                                                           +  182.25')
        exp_val.append('|                                                           |')
        exp_val.append('|                                o                          |')
        exp_val.append('|                          o                         *      *')
        exp_val.append('|                   o            *      *      *            |')
        exp_val.append('o------o-----o-|----*------*--|--------------|--------------|  0.00')
        exp_val.append('0.00e+00.......2.25e+00.......4.50e+00.......6.75e+00.......9.00e+00')
        self.assertListEqual(exp_val, act_val)

    def test_TextPlot_dunder_str_dunder(self):
        self.maxDiff = None
        # Prepare test data
        ncur=2
        npts=10
        x=[array('f'),array('f')]
        y=[array('f'),array('f')]
        for i in range(npts):
            x[0].append(float(i))
            x[1].append(float(i))
            y[0].append(float(i*i))
            y[1].append(float(i*i*i))
        symbol=array('u')
        symbol.append('*')
        symbol.append('o')
        titl1='Test ASCII Plot'
        titl2='*: y = x^2; o: y = x^3'
        plot=TextPlot(ncur,npts,x,y,symbol,titl1,titl2)
        act_val = str(plot)
        print(f"Actual Value:")
        print(f"{act_val}")
        exp_val=''
        exp_val+='Test ASCII Plot\n'
        exp_val+='*: y = x^2; o: y = x^3\n'
        exp_val+=' \n'
        exp_val+='0.00e+00.......2.25e+00.......4.50e+00.......6.75e+00.......9.00e+00\n'
        exp_val+='|--------------|--------------|--------------|--------------o  729.00\n'
        exp_val+='|                                                           |\n'
        exp_val+='|                                                           |\n'
        exp_val+='|                                                           |\n'
        exp_val+='|                                                           |\n'
        exp_val+='+                                                           +  546.75\n'
        exp_val+='|                                                    o      |\n'
        exp_val+='|                                                           |\n'
        exp_val+='|                                                           |\n'
        exp_val+='|                                                           |\n'
        exp_val+='+                                                           +  364.50\n'
        exp_val+='|                                              o            |\n'
        exp_val+='|                                                           |\n'
        exp_val+='|                                                           |\n'
        exp_val+='|                                       o                   |\n'
        exp_val+='+                                                           +  182.25\n'
        exp_val+='|                                                           |\n'
        exp_val+='|                                o                          |\n'
        exp_val+='|                          o                         *      *\n'
        exp_val+='|                   o            *      *      *            |\n'
        exp_val+='o------o-----o-|----*------*--|--------------|--------------|  0.00\n'
        exp_val+='0.00e+00.......2.25e+00.......4.50e+00.......6.75e+00.......9.00e+00\n'
        print(f"Expected Value:")
        print(f"{exp_val}")
        self.assertEqual(exp_val, act_val)

    def test_qplot_function(self):
        self.maxDiff = None
        # Redirect standard output to a buffer
        captured_output = StringIO()
        sys.stdout = captured_output
        # Prepare test data
        ncur=2
        npts=10
        x=[array('f'),array('f')]
        y=[array('f'),array('f')]
        for i in range(npts):
            x[0].append(float(i))
            x[1].append(float(i))
            y[0].append(float(i*i))
            y[1].append(float(i*i*i))
        symbol=array('u')
        symbol.append('*')
        symbol.append('o')
        titl1='Test ASCII Plot'
        titl2='*: y = x^2; o: y = x^3'
        # Prepare expected output
        exp_val=''
        exp_val+='Test ASCII Plot\n'
        exp_val+='*: y = x^2; o: y = x^3\n'
        exp_val+=' \n'
        exp_val+='0.00e+00.......2.25e+00.......4.50e+00.......6.75e+00.......9.00e+00\n'
        exp_val+='|--------------|--------------|--------------|--------------o  729.00\n'
        exp_val+='|                                                           |\n'
        exp_val+='|                                                           |\n'
        exp_val+='|                                                           |\n'
        exp_val+='|                                                           |\n'
        exp_val+='+                                                           +  546.75\n'
        exp_val+='|                                                    o      |\n'
        exp_val+='|                                                           |\n'
        exp_val+='|                                                           |\n'
        exp_val+='|                                                           |\n'
        exp_val+='+                                                           +  364.50\n'
        exp_val+='|                                              o            |\n'
        exp_val+='|                                                           |\n'
        exp_val+='|                                                           |\n'
        exp_val+='|                                       o                   |\n'
        exp_val+='+                                                           +  182.25\n'
        exp_val+='|                                                           |\n'
        exp_val+='|                                o                          |\n'
        exp_val+='|                          o                         *      *\n'
        exp_val+='|                   o            *      *      *            |\n'
        exp_val+='o------o-----o-|----*------*--|--------------|--------------|  0.00\n'
        exp_val+='0.00e+00.......2.25e+00.......4.50e+00.......6.75e+00.......9.00e+00\n'
        # Call the qplot function, which prints to stdout
        qplot(ncur,npts,x,y,symbol,titl1,titl2)
        # Get the captured output
        act_val = captured_output.getvalue().strip()+'\n'
        # Reset the standard output
        sys.stdout = sys.__stdout__
        # Compare expected and actual output
        self.assertEqual(exp_val, act_val)


if __name__ == '__main__':
    unittest.main()



"""
This module contains unit tests for the Strand, FENEStrand, and FENSStrand classes, as well as the beta function.
"""


# Standard imports
import unittest

# Local imports
from RheologyNetworkModelSimulator.strand import Strand, FENEStrand, beta, FENSStrand


class Test_beta(unittest.TestCase):
    def test_beta(self):
        z, w = 1.5, 2.5
        exp_val = 0.1963495  # Calculated by R
        act_val = beta(z, w)
        self.assertAlmostEqual(exp_val, act_val)


class Test_Strand(unittest.TestCase):
    def test_init(self):
        s = Strand(0.1, 0.2, 0.3)
        exp_val = (0.1, 0.2, 0.3)
        act_val = (s.qx, s.qy, s.qz)
        self.assertTupleEqual(exp_val, act_val)

    def test_property_getter_max_length(self):
        exp_val = None
        s = Strand(0.1, 0.2, 0.3)
        act_val = s.max_length
        self.assertEqual(exp_val, act_val)

    def test_property_getters_qx_qy_qz(self):
        exp_val = (0.1, 0.2, 0.3)
        s = Strand(0.1, 0.2, 0.3)
        act_val = (s.qx, s.qy, s.qz)
        self.assertTupleEqual(exp_val, act_val)

    def test_property_setter_qx(self):
        exp_val = 0.4
        s = Strand(0.1, 0.2, 0.3)
        s.qx = 0.4
        act_val = s.qx
        self.assertEqual(exp_val, act_val)

    def test_property_setter_qy(self):
        exp_val = 0.4
        s = Strand(0.1, 0.2, 0.3)
        s.qy = 0.4
        act_val = s.qy
        self.assertEqual(exp_val, act_val)

    def test_property_setter_qz(self):
        exp_val = 0.4
        s = Strand(0.1, 0.2, 0.3)
        s.qz = 0.4
        act_val = s.qz
        self.assertEqual(exp_val, act_val)

    def test_get_qs(self):
        exp_val = (0.1, 0.2, 0.3)
        s = Strand(0.1, 0.2, 0.3)
        act_val = s.get_qs()
        self.assertTupleEqual(exp_val, act_val)

    def test_str_len_sqr(self):
        exp_val = 0.14
        s = Strand(0.1, 0.2, 0.3)
        act_val = s.str_len_sqr()
        self.assertEqual(exp_val, act_val)

    def test_stress_XX(self):
        s = Strand(0.1, 0.2, 0.3)
        self.assertRaises(NotImplementedError, s.stress_XX)

    def test_stress_YY(self):
        s = Strand(0.1, 0.2, 0.3)
        self.assertRaises(NotImplementedError, s.stress_YY)

    def test_stress_ZZ(self):
        s = Strand(0.1, 0.2, 0.3)
        self.assertRaises(NotImplementedError, s.stress_ZZ)

    def test_stress_YX(self):
        s = Strand(0.1, 0.2, 0.3)
        self.assertRaises(NotImplementedError, s.stress_YX)

    def test_loss_rate(self):
        s = Strand(0.1, 0.2, 0.3)
        self.assertRaises(NotImplementedError, s.loss_rate)

    def test_loss_prob(self):
        s = Strand(0.1, 0.2, 0.3)
        self.assertRaises(NotImplementedError, s.loss_prob, 0.001)

    def test_generate_eq_ensemble(self):
        s = Strand(0.1, 0.2, 0.3)
        self.assertRaises(NotImplementedError, s.generate_eq_ensemble, 10)


class Test_FENEStrand(unittest.TestCase):
    def test_init(self):
        s = FENEStrand(qx=0.1, qy=0.2, qz=0.3, b=100.0, n=2.0, mm=1.0)
        exp_val = (0.1, 0.2, 0.3)
        act_val = (s.qx, s.qy, s.qz)
        self.assertTupleEqual(exp_val, act_val)
        self.assertEqual(100.0, s.b)
        self.assertEqual(2.0, s.n)
        self.assertEqual(1.0, s.mm)

    def test_strand_init_to_long(self):
        self.assertRaises(AssertionError, FENEStrand, 1.0, 1.0, 1.0)

    def test_property_getter_max_length(self):
        exp_val = 1.0
        s = FENEStrand()
        act_val = s.max_length
        self.assertEqual(exp_val, act_val)

    def test_property_setter_qx_bad(self):
        s = FENEStrand(0.1, 0.2, 0.3)
        self.assertRaises(AssertionError, setattr, s, 'qx', 0.94)

    def test_property_setter_qy_bad(self):
        s = FENEStrand(0.1, 0.2, 0.3)
        self.assertRaises(AssertionError, setattr, s, 'qy', 0.95)

    def test_property_setter_qz_bad(self):
        s = FENEStrand(0.1, 0.2, 0.3)
        self.assertRaises(AssertionError, setattr, s, 'qz', 0.98)

    def test_stress_XX(self):
        exp_val = 1.1627906976744187
        s = FENEStrand(0.1, 0.2, 0.3)
        act_val = s.stress_XX()
        self.assertEqual(exp_val, act_val)

    def test_stress_YY(self):
        exp_val = 4.651162790697675
        s = FENEStrand(0.1, 0.2, 0.3)
        act_val = s.stress_YY()
        self.assertEqual(exp_val, act_val)

    def test_stress_ZZ(self):
        exp_val = 10.465116279069768
        s = FENEStrand(0.1, 0.2, 0.3)
        act_val = s.stress_ZZ()
        self.assertEqual(exp_val, act_val)

    def test_stress_YX(self):
        exp_val = 2.3255813953488373
        s = FENEStrand(0.1, 0.2, 0.3)
        act_val = s.stress_YX()
        self.assertEqual(exp_val, act_val)

    def test_loss_rate(self):
        exp_val = 1.1627906976744187
        s = FENEStrand(0.1, 0.2, 0.3)
        act_val = s.loss_rate()
        self.assertEqual(exp_val, act_val)

    def test_loss_prob(self):
        exp_val = 0.001162114918526358
        s = FENEStrand(0.1, 0.2, 0.3)
        act_val = s.loss_prob(0.001)
        self.assertEqual(exp_val, act_val)

    def test_generate_eq_ensemble(self):
        from random import seed
        exp_val = (0.1314922734575693, 0.03212588679044972, 0.029485861493129577)
        seed(1234567890)
        e = FENEStrand().generate_eq_ensemble(10)
        self.assertEqual(10, len(e))
        act_val = (e[0].qx, e[0].qy, e[0].qz)
        self.assertTupleEqual(exp_val, act_val)


class Test_FENSStrand(unittest.TestCase):
    def test_init(self):
        s = FENSStrand(qx=0.1, qy=0.2, qz=0.3)
        exp_val = (0.1, 0.2, 0.3)
        act_val = (s.qx, s.qy, s.qz)
        self.assertTupleEqual(exp_val, act_val)

    def test_strand_init_to_long(self):
        self.assertRaises(AssertionError, FENSStrand, 100.0, 100.0, 100.0)

    def test_property_getter_max_length(self):
        exp_val = 100.0
        s = FENSStrand()
        act_val = s.max_length
        self.assertEqual(exp_val, act_val)

    def test_property_setter_qx_bad(self):
        s = FENSStrand(10.0, 20.0, 30.0)
        self.assertRaises(AssertionError, setattr, s, 'qx', 94.0)

    def test_property_setter_qy_bad(self):
        s = FENSStrand(10.0, 20.0, 30.0)
        self.assertRaises(AssertionError, setattr, s, 'qy', 95.0)

    def test_property_setter_qz_bad(self):
        s = FENSStrand(10.0, 20.0, 30.0)
        self.assertRaises(AssertionError, setattr, s, 'qz', 98.0)

    def test_stress_XX(self):
        exp_val = 0.1*0.1
        s = FENSStrand(0.1, 0.2, 0.3)
        act_val = s.stress_XX()
        self.assertEqual(exp_val, act_val)

    def test_stress_YY(self):
        exp_val = 0.2*0.2
        s = FENSStrand(0.1, 0.2, 0.3)
        act_val = s.stress_YY()
        self.assertEqual(exp_val, act_val)

    def test_stress_ZZ(self):
        exp_val = 0.3*0.3
        s = FENSStrand(0.1, 0.2, 0.3)
        act_val = s.stress_ZZ()
        self.assertEqual(exp_val, act_val)

    def test_stress_YX(self):
        exp_val = 0.1*0.2
        s = FENSStrand(0.1, 0.2, 0.3)
        act_val = s.stress_YX()
        self.assertEqual(exp_val, act_val)

    def test_loss_rate(self):
        exp_val = 1.0
        s = FENSStrand(0.1, 0.2, 0.3)
        act_val = s.loss_rate()
        self.assertEqual(exp_val, act_val)

    def test_loss_prob(self):
        exp_val = 0.001
        s = FENSStrand(0.1, 0.2, 0.3)
        act_val = s.loss_prob(0.001)
        # TODO: Investigate why this needs almost equal
        self.assertAlmostEqual(exp_val, act_val, 5)

    def test_generate_eq_ensemble(self):
        from random import seed
        exp_val = (-0.2686617612333387, 0.44611821880006447, 0.17019087101967528)
        seed(1234567890)
        e = FENSStrand().generate_eq_ensemble(10)
        self.assertEqual(10, len(e))
        act_val = (e[0].qx, e[0].qy, e[0].qz)
        self.assertTupleEqual(exp_val, act_val)


if __name__ == '__main__':
    unittest.main()



"""
This module defines the class Strand, and its subclasses, which represents polymer strands in a network model
for polymer melt rheology.

Exported classes:
    Strand: Base class to represent a strand in a polymer melt network model.
    FENSStrand: Subclass of Strand to represent a FENS (finitely extensible network strand) strand in a polymer melt network model.
    FENEStrand: Subclass of Strand to represent a FENE (finitely extensible non-linear elastic) strand in a polymer melt network model.

Exported functions:
    beta: Function to compute the mathematical beta function.

Exported exceptions:
    None
"""


# Standard imports
from math import pow, sqrt, exp, lgamma, pi, sin, cos, log
from random import uniform


class Strand:
    """
    Represent a strand in a network model for polymer melt rheology.

    Reference: Geurts, K.R. and L.E. Wedgewood, "A finitely extensible network strand model with nonlinear backbone
    forces and entanglement kinetics," J. Chem. Phys., 1-January-1997, 106(1), pp. 339-346.

    Nondimensionalization Scheme:
        Strand internal coordinates are divided by Qo, the maximum strand length ??Hookean has no max length??.
        Loss rate is multiplied by the loss rate constant, Lambdao.
        Time step size is divided by the loss rate constant, Lamdao.
        Stress tensor is divided by NokT.
    """
    def __init__(self, qx=0.0, qy=0.0, qz=0.0):
        """
        Construct a Strand with given X, Y, and Z internal coordinate lengths, all non-dimensionalized by
        dividing by the maximum strand length Qo.

        :param qx: Internal X-coordinate of Strand length, non-dimensionalizec by dividing by Qo, the maximum strand length, as float
        :param qy: Internal Y-coordinate of Strand length, non-dimensionalizec by dividing by Qo, the maximum strand length, as float
        :param qz: Internal Z-coordinate of Strand length, non-dimensionalizec by dividing by Qo, the maximum strand length, as float
        """
        self._max_length = None
        self._qx=qx
        self._qy=qy
        self._qz=qz

    @property
    def max_length(self):
        return self._max_length

    @property
    def qx(self):
        return self._qx

    # TODO: The setters for the internal strand coordinates enforce the length constraint immediately.
    # But be aware that since they can be called independently, there could be a situation where the user
    # plans to change the other coordinates as well, so that the final length is valid, but when the first to second
    # setter is called, that it isn't valid yet.

    @qx.setter
    def qx(self, value):
        if self._max_length is not None:
            assert(sqrt(value*value + self.qy*self.qy + self.qz*self.qz) <= self._max_length)
        self._qx = value

    @property
    def qy(self):
        return self._qy

    @qy.setter
    def qy(self, value):
        if self._max_length is not None:
            assert(sqrt(self.qx*self.qx + value*value + self.qz*self.qz) <= self._max_length)
        self._qy = value

    @property
    def qz(self):
        return self._qz

    @qz.setter
    def qz(self, value):
        if self._max_length is not None:
            assert(sqrt(self.qx*self.qx + self.qy*self.qy + value*value) <= self._max_length)
        self._qz = value

    def get_qs(self):
        """
        Get the internal coordinates of the Strand, all non-dimensionalized by dividing by the maximum strand length Qo.
        :return: Tuple (qx, qy, qz) of internal coordinates of the Strand, as (float, float, float)
        """
        return (self.qx, self.qy, self.qz)

    def str_len_sqr(self):
        """
        Compute the length of the Strand, squared. Non-dimensionalized by dividing by maximum strand length Qo, squared.
        :return: Strand length squared, as float
        """
        return pow(self.qx, 2)+pow(self.qy, 2)+pow(self.qz, 2)

    def loss_rate(self):
        """
        Compute the loss_rate for the Strand. Loss rate is multiplied by the loss rate constant, Lambdao.
        Must be implemented by subclasses.
        :return: Loss rate for strand, multiplied by the loss rate constant, Lambdao, as float
        """
        raise NotImplementedError("Subclasses must implement loss_rate method.")
        return None

    def loss_prob(self, eps):
        """
        Compute the probability of the Strand disentangling from the network during a time step of length eps.
        :param eps: Time step size, divided by the loss rate constant, Lamdao, as float
        :return: Probability of the Strand disentangling from the network during a time step of length eps, as float
        """
        return 1.0 - exp(-1.0 * self.loss_rate() * eps)

    # XX component of total stress tensor
    def stress_XX(self):
        """
        Compute the XX-component of the strand's contrribution to the total stress tensor of the network.
        Stress tensor is divided by NokT. Must be implemented by subclasses.
        :return: XX-component of the strand's contribution to the total stress tensor of the network, as float
        """
        raise NotImplementedError("Subclasses must implement stress_XX method.")
        return None

    # YY component of total stress tensor
    def stress_YY(self):
        """
        Compute the YY-component of the strand's contribution to the total stress tensor of the network.
        Stress tensor is divided by NokT. Must be implemented by subclasses.
        :return: YY-component of the strand's contrribution to the total stress tensor of the network, as float
        """
        raise NotImplementedError("Subclasses must implement stress_YY method.")
        return None

    # ZZ component of total stress tensor
    def stress_ZZ(self):
        """
        Compute the ZZ-component of the strand's contribution to the total stress tensor of the network.
        Stress tensor is divided by NokT. Must be implemented by subclasses.
        :return: ZZ-component of the strand's contrribution to the total stress tensor of the network, as float
        """
        raise NotImplementedError("Subclasses must implement stress_ZZ method.")
        return None

    # YX component of total stress tensor
    def stress_YX(self):
        """
        Compute the YX-component of the strand's contrribution to the total stress tensor of the network.
        Stress tensor is divided by NokT. Must be implemented by subclasses.
        :return: YX-component of the strand's contrribution to the total stress tensor of the network, as float
        """
        raise NotImplementedError("Subclasses must implement stress_YX method.")
        return None

    def generate_eq_ensemble(self, n=1):
        """
        Create an ensemble of size n number of strands, where the distribution of strand lengths will
        fit the equilibrium distribution. Must be implemented by subclasses.
        :param n: Number of strands to generate for the ensemble, as int
        :return: List of strands in the equilibrium ensemble, as [Strand objects]
        """
        raise NotImplementedError("Subclasses must implement generate_eq_ensemble method.")
        return None


class FENSStrand(Strand):
    """
    Represent a FENS (finitely extensible network strand) strand in a network model for polymer melt rheology.

    References:
    
    (1) Wedgewood, L.E. and K.R. Geurts, Rheol. Acta 34, 196 (1995).
    (2) Geurts, K.R. and L.E. Wedgewood, "A finitely extensible network strand model with nonlinear backbone
        forces and entanglement kinetics," J. Chem. Phys., 1-January-1997, 106(1), pp. 339-346.

    Non-affine Motion prefactor is  1-(Q/Qo)^n. No strand may exceed the nondimensional length of one.
    The Hookean (linear) force law is used. The loss rate is constant.

    Nondimensionalization Scheme:
        Strand internal coordinates are divided by sqrt(kT/H).
        Loss rate is multiplied by the loss rate constant, Lambdao.
        Time step size is divided by the loss rate constant, Lamdao.
        Stress tensor is divided by NokT.

    NOTE: For the FENS Strand, the non-dimensionalization scheme is different than for the FENE Strand and the base Strand class.
    This difference is consistent with the two journal articles that define these to strand models.
    And it should be noted that for the FENS Strand, b is always = 100.0.
    This can be rationalized through the relationship that q{n-dim-fene} = q/qo = q/bsqrt(kT/H) = q{n-dim-fens}/b.
    The way this manifests in the code is that FENEStrands._max_length = 1.0, and FENSStrands._max_length = 100.0.
    This is immaterial for comparing most results, although it does mean that direct comparison of strand length values
    between the models is not meaninful without using the above formula to place them on a consisten basis.
    A possible TODO is to refactor the code to have a consistent non-dimensionalization scheme across the Strand subclasses.
    For now, I will opt to keep the non-dimensionalization scheme as it is, since it is consistent with the journal articles that
    define these two strand models.
    """
    def __init__(self, qx=0.0, qy=0.0, qz=0.0):
        """
        Construct a Strand with given X, Y, and Z internal coordinate lengths, all non-dimensionalized by
        dividing by sqrt(kT/H).
        :param qx: Internal X-coordinate of Strand length, non-dimensionalizec by dividing by sqrt(kT/H), as float
        :param qy: Internal Y-coordinate of Strand length, non-dimensionalizec by dividing by sqrt(kT/H), as float
        :param qz: Internal Z-coordinate of Strand length, non-dimensionalizec by dividing by sqrt(kT/H), as float
        """
        super().__init__(qx, qy, qz)
        # Override max length to 100.0 for FENS Strand. Must be >> 3.0. Non-dimensionalization by dividing by sqrt(kT/H).
        self._max_length = 100.0 
        assert(sqrt(qx*qx + qy*qy + qz*qz) <= self._max_length)

    def loss_rate(self):
        """
        Compute the loss_rate for the Strand. Loss rate is multiplied by the loss rate constant, Lambdao.
        :return: Loss rate for strand, multiplied by the loss rate constant, Lambdao, as float
        """
        return 1.0

    # XX component of total stress tensor
    def stress_XX(self):
        """
        Compute the XX-component of the strand's contrribution to the total stress tensor of the network.
        Stress tensor is divided by NokT.
        :return: XX-component of the strand's contrribution to the total stress tensor of the network, as float
        """
        return self.qx * self.qx

    # YY component of total stress tensor
    def stress_YY(self):
        """
        Compute the YY-component of the strand's contrribution to the total stress tensor of the network.
        Stress tensor is divided by NokT.
        :return: YY-component of the strand's contrribution to the total stress tensor of the network, as float
        """
        return self.qy * self.qy

    # ZZ component of total stress tensor
    def stress_ZZ(self):
        """
        Compute the ZZ-component of the strand's contrribution to the total stress tensor of the network.
        Stress tensor is divided by NokT.
        :return: ZZ-component of the strand's contrribution to the total stress tensor of the network, as float
        """
        return self.qz * self.qz

    # YX component of total stress tensor
    def stress_YX(self):
        """
        Compute the YX-component of the strand's contrribution to the total stress tensor of the network.
        Stress tensor is divided by NokT.
        :return: YX-component of the strand's contrribution to the total stress tensor of the network, as float
        """
        return self.qy * self.qx

    def generate_eq_ensemble(self, n=1):
        """
        Create an ensemble of size n number of strands, where the distribution of strand lengths will
        fit the equilibrium distribution.
        :param n: Number of strands to generate for the ensemble, as int
        :return: List of strands in the equilibrium ensemble, as [Strand objects]
        """
        e = []
        for s in range(n):
            # _qx:
            p1 = uniform(0.0,1.0)
            if p1 == 0.0: p1 = 1.0
            p2 = uniform(0.0,1.0)
            _qx = sqrt(-2.0 * log(p1)) * cos(2.0 * pi * p2)
            # _qy:
            p1 = uniform(0.0,1.0)
            if p1 == 0.0: p1 = 1.0
            p2 = uniform(0.0,1.0)
            _qy = sqrt(-2.0 * log(p1)) * cos(2.0 * pi * p2)
            # _qz:
            p1 = uniform(0.0,1.0)
            if p1 == 0.0: p1 = 1.0
            p2 = uniform(0.0,1.0)
            _qz = sqrt(-2.0 * log(p1)) * cos(2.0 * pi * p2)
            if self.str_len_sqr() <= self.max_length * self.max_length:
                # Strand is of legal length, so add to ensemble
                e.append(FENSStrand(_qx, _qy, _qz))
            else:
                # TODO: Raise a specific exception instead
                assert(False, "Generated strand of length greater than max length. This should be very unlikely, but if it happend!")
        return e


class FENEStrand(Strand):
    """
    Represent a FENE (finitely extensible non-linear elastic) strand in a network model for polymer melt rheology.

    Reference: Geurts, K.R. and L.E. Wedgewood, "A finitely extensible network strand model with nonlinear backbone
    forces and entanglement kinetics," J. Chem. Phys., 1-January-1997, 106(1), pp. 339-346.

    Non-affine Motion prefactor is  1-(Q/Qo)^n. No strand may exceed the nondimensional length of one.
    The FENE force law is used. The loss rate is (1/Lambdao)(1)/(1-(Q/Qo)^n*m).

    Nondimensionalization Scheme:
        Strand internal coordinates are divided by Qo, the maximum strand length.
        Loss rate is multiplied by the loss rate constant, Lambdao.
        Time step size is divided by the loss rate constant, Lamdao.
        Stress tensor is divided by NokT.
    """
    def __init__(self, qx=0.0, qy=0.0, qz=0.0, b=100.0, n=2.0, mm=1.0):
        """
        Construct a Strand with given X, Y, and Z internal coordinate lengths, all non-dimensionalized by
        dividing by the maximum strand length Qo.
        :param qx: Internal X-coordinate of Strand length, non-dimensionalizec by dividing by Qo, the maximum strand length, as float
        :param qy: Internal Y-coordinate of Strand length, non-dimensionalizec by dividing by Qo, the maximum strand length, as float
        :param qz: Internal Z-coordinate of Strand length, non-dimensionalizec by dividing by Qo, the maximum strand length, as float
        :param b: The FENE Parameter (HQo^2/kT), or non-dimensional maximum strand length, as float        
        :param n: Exponent on Q/Qo in the non-affine motion prefactor. Should be equal to 2, as float
        :param mm: mm*n is exponent on (Q/Qo) in the loss rate. mm should be equal to 1, as float
        """
        super().__init__(qx, qy, qz)
        self._max_length = 1.0 # Override max length to 1.0 for FENE Strand
        assert(sqrt(qx*qx + qy*qy + qz*qz) <= self._max_length)
        assert(b>0.0)
        assert(n==2.0)
        assert(mm==1.0)
        self._b = b
        self._n = n
        self._mm = mm

    @property
    def b(self):
        return self._b

    @property
    def n(self):
        return self._n

    @property
    def mm(self):
        return self._mm

    def loss_rate(self):
        """
        Compute the loss_rate for the Strand. Loss rate is multiplied by the loss rate constant, Lambdao.
        :return: Loss rate for strand, multiplied by the loss rate constant, Lambdao, as float
        """
        return 1.0 / (1.0 - pow(sqrt(self.str_len_sqr()), (self.mm * self.n)))

    # XX component of total stress tensor
    def stress_XX(self):
        """
        Compute the XX-component of the strand's contrribution to the total stress tensor of the network.
        Stress tensor is divided by NokT.
        :return: XX-component of the strand's contrribution to the total stress tensor of the network, as float
        """
        return self.b * self.qx * self.qx / (1.0 - self.str_len_sqr())

    # YY component of total stress tensor
    def stress_YY(self):
        """
        Compute the YY-component of the strand's contrribution to the total stress tensor of the network.
        Stress tensor is divided by NokT.
        :return: YY-component of the strand's contrribution to the total stress tensor of the network, as float
        """
        return self.b * self.qy * self.qy / (1.0 - self.str_len_sqr())

    # ZZ component of total stress tensor
    def stress_ZZ(self):
        """
        Compute the ZZ-component of the strand's contrribution to the total stress tensor of the network.
        Stress tensor is divided by NokT.
        :return: ZZ-component of the strand's contrribution to the total stress tensor of the network, as float
        """
        return self.b * self.qz * self.qz / (1.0 - self.str_len_sqr())

    # YX component of total stress tensor
    def stress_YX(self):
        """
        Compute the YX-component of the strand's contrribution to the total stress tensor of the network.
        Stress tensor is divided by NokT.
        :return: YX-component of the strand's contrribution to the total stress tensor of the network, as float
        """
        return self.b * self.qy * self.qx / (1.0 - self.str_len_sqr())

    def generate_eq_ensemble(self, n=1):
        """
        Create an ensemble of size n number of strands, where the distribution of strand lengths will
        fit the equilibrium distribution.
        :param n: Number of strands to generate for the ensemble, as int
        :return: List of strands in the equilibrium ensemble, as [Strand objects]
        """
        e = []
        # jeq: Normalization constant for equilibriium FENE Dumbbell distribution function DPL2 Eq. (L) of Table 11.5-1
        jeq = (1. / (2. * pi * beta(3. / 2., (self.b + 2.) / 2.)))
        for s in range(n):
            p1 = 1.0
            p2 = 0.0
            # r, theta, phi: Internal coordinates of a network strand in spherical coordinates used to find equilibrium distribution.
            r, theta, phi = 0.0, 0.0, 0.0
            while p1>p2:
                r = uniform(0.0,1.0)
                theta = pi * uniform(0.0, 1.0)
                phi = 2.0 * pi * uniform(0.0, 1.0)
                p1 = uniform(0.0, 1.0)
                p2 = jeq * ((1.0 - r * r) ** (self.b / 2.0)) * r * r * sin(theta)
            x = r * sin(theta) * cos(phi)
            y = r * sin(theta) * sin(phi)
            z = r * cos(theta)
            e.append(FENEStrand(x, y, z, self.b, self.n, self.mm))
        return e


def beta(z, w):
    """
    Compute beta function of z and w.

    :param z: Parameter z of beta function, as float
    :param w: Parameter w of beta function, as float
    :return: Beta function of z and w, as float
    """
    return exp(lgamma(z)+lgamma(w)-lgamma(z+w))


import unittest

from RheologyNetworkModelSimulator.ensemble import Ensemble
from RheologyNetworkModelSimulator.strand import Strand, FENEStrand

class Test_Ensemble(unittest.TestCase):
    def test_q_ave(self):
        exp_val = 0.37416573867739417
        e = Ensemble([Strand(0.1, 0.2, 0.3), Strand(0.1, 0.2, 0.3)])
        act_val = e.q_ave()
        self.assertEqual(exp_val, act_val)

    def test_ensemble_stress(self):
        exp_val = (2.*1.1627906976744187, 2.*4.651162790697675, 2.*10.465116279069768, 2.*2.3255813953488373)
        e=Ensemble([FENEStrand(0.1, 0.2, 0.3), FENEStrand(0.1, 0.2, 0.3)])
        act_val = e.ensemble_stress()
        self.assertTupleEqual(exp_val, act_val)

    def test_ensemble_stress_fail(self):
        exp_val = (2.*1.1627906976744187, 2.*4.651162790697675, 2.*10.465116279069768, 2.*2.3255813953488373)
        e=Ensemble([Strand(0.1, 0.2, 0.3), Strand(0.1, 0.2, 0.3)])
        self.assertRaises(NotImplementedError, e.ensemble_stress)

    def test_len(self):
        e = Ensemble([Strand(0.1, 0.2, 0.3), Strand(0.1, 0.2, 0.3)])
        exp_val = 2
        act_val = len(e)
        self.assertEqual(exp_val, act_val)

    def test_getitem(self):
        e = Ensemble([Strand(0.1, 0.2, 0.3), Strand(0.2, 0.3, 0.1)])
        exp_val = (0.2, 0.3, 0.1)
        act_val = e[1].get_qs()
        self.assertTupleEqual(exp_val, act_val)

    def test_iter(self):
        e = Ensemble([Strand(0.1, 0.2, 0.3), Strand(0.2, 0.3, 0.1)])
        qs_list = []
        for s in e:
            qs_list.append(s.get_qs())
        exp_val = [(0.1, 0.2, 0.3), (0.2, 0.3, 0.1)]
        self.assertListEqual(exp_val, qs_list)

    def test_remove(self):
        e = Ensemble([Strand(0.1, 0.2, 0.3), Strand(0.2, 0.3, 0.1)])
        s_to_remove = e[1]
        e.remove(s_to_remove)
        exp_val = 1
        act_val = len(e)
        self.assertEqual(exp_val, act_val)
        exp_qs = (0.1, 0.2, 0.3)
        act_qs = e[0].get_qs()
        self.assertTupleEqual(exp_qs, act_qs)

    def test_append(self):
        e = Ensemble([Strand(0.1, 0.2, 0.3), Strand(0.2, 0.3, 0.1)])
        new_strand = Strand(0.3, 0.1, 0.2)
        e.append(new_strand)
        exp_val = 3
        act_val = len(e)
        self.assertEqual(exp_val, act_val)
        exp_qs = (0.3, 0.1, 0.2)
        act_qs = e[2].get_qs()
        self.assertTupleEqual(exp_qs, act_qs)

    def test_property_strands(self):
        exp_val = [Strand(0.1, 0.2, 0.3), Strand(0.2, 0.3, 0.1)]
        e = Ensemble(exp_val)
        act_val = e.strands
        self.assertListEqual(exp_val, act_val)
        # Remove a strand from the copied list returned by the property and check that the ensemble is unchanged
        act_val.remove(act_val[0])
        self.assertTrue(len(e) == 2) # Length is still 2
        self.assertListEqual(exp_val, e.strands) # Strands are unchanged


if __name__ == '__main__':
    unittest.main()



"""
This module defines the class TextPlot, which represents a simple ASCII scatter plot.

Exported classes:
    TextPlot: Class to create a simple ASCII scatter plot.

Exported functions:
    qplot: Function to print a simple ASCII plot to stdout.

Exported exceptions:
    None
"""

# stadard imports
from array import array


# TODO: (1) Add axis labels that print as a title line of the form: 'x (units) vs y (units)'
# TODO: (2) Add a plot legend, with one string per curve, showing the symbol and a description of the curve, to be
# printed as a title line.
class TextPlot():
    """
    Class to create a simple ASCII plot.
    Usage:
        (1) Create an instance of the TextPlot class with the parameters that define the plot.
        (2) Use the str() function to get the ASCII plot as a string.
        (3) Print the string or write it to a file as desired.
    """
    def __init__(self, ncur=1, npts=10, x=None, y=None, symbol=None, titl1='', titl2=''):
        """
        Provide all of the data needed to create the ASCII plot.
        :parameter ncur: Number of curves to plot, int
        :parmaeter npts: Number of points per curve, int
        :parameter x: X-coordinates of points, list of ncur arrays of npts floats, [array('f') ... array('f')]
        :parameter y: Y-coordinates of points, list of ncur arrays of npts floats, [array('f') ... array('f')]
        :parameter symbol: Plotting symbol for each curve, array of ncur characters, array('u')
        :parameter titl1: Title line 1, string
        :parameter titl2: Title line 2, string
        """
        assert(int(ncur)>0)
        self._ncur=int(ncur)
        assert(int(npts)>0)
        self._npts=int(npts)
        assert(type(x)==list)
        # TODO: Strictly speaking, if the code in _qplot() is changed, not all curves would have to have the same number of points.
        # Alhough the number of x-points and y-points for each curve would still have to match.
        for curve in x:
            assert(type(curve)==array)
            assert(len(curve)==int(self._npts))
        self._x=x
        assert(type(y)==list)
        for curve in y:
            assert(type(curve)==array)
            assert(len(curve)==int(self._npts))
        self._y=y
        assert(type(symbol)==array)
        assert(len(symbol)==int(self._ncur))
        self._symbol=symbol
        self._titl1=str(titl1)
        self._titl2=str(titl2)

    def _qplot(self):
        """
        Utility function called by __str__() to build the list of strings that make up the plot.
        :return: list of strings that make up the plot
        """
        # Characters for line of the plot
        cpl = 60
        # Lines (not curves, but ASCII lines) per plot
        lpp = 20

        # Set up the plot array
        line=[]
        # First, blank out the entire plot area with ' ' including the borders
        for i in range(lpp+1): # 0..lpp inclusive
            a_line = array('u')
            for j in range(cpl+1): # 0..cpl inclusive
                a_line.append(' ')
            line.append(a_line)
        # Now create top and bottome axes
        for j in range(cpl+1):
            line[0][j]='-'
            line[lpp][j]='-'
        # Now create x tick marks
        line[0][round(cpl/4)]='|'
        line[lpp][round(cpl/4)]='|'
        line[0][round(cpl/2)]='|'
        line[lpp][round(cpl/2)]='|'
        line[0][round(3*cpl/4)]='|'
        line[lpp][round(3*cpl/4)]='|'
        # Now create left and right axes
        for i in range(lpp+1):
            line[i][0]='|'
            line[i][cpl]='|'
        # Now create y tick marks
        line[round(lpp/4)][0]='+'
        line[round(lpp/4)][cpl]='+'
        line[round(lpp/2)][0]='+'
        line[round(lpp/2)][cpl]='+'
        line[round(3*lpp/4)][0]='+'
        line[round(3*lpp/4)][cpl]='+'

        # Find maximum and minimum values of x and y accross all curves, for scaling
        # TODO: May be a more Pythonic way to do this
        xmax=self._x[0][0]
        xmin=self._x[0][0]
        ymax=self._y[0][0]
        ymin=self._y[0][0]
        for j in range(self._ncur):
            for i in range(self._npts):
                if self._x[j][i]>xmax:
                    xmax=self._x[j][i]
                if self._x[j][i]<xmin:
                    xmin=self._x[j][i]
                if self._y[j][i]>ymax:
                    ymax=self._y[j][i]
                if self._y[j][i]<ymin:
                    ymin=self._y[j][i]

        # Fill in the points
        for j in range(self._ncur):
            for i in range(self._npts):
                yi=round((lpp)*(self._y[j][i]-ymin)/(ymax-ymin))
                xi=round((cpl)*(self._x[j][i]-xmin)/(xmax-xmin))
                line[yi][xi]=self._symbol[j]

        # Now build the list of strings that make up the plot

        plot_strings=[]
                
        # Add the titles
        plot_strings.append(self._titl1)
        plot_strings.append(self._titl2)
        plot_strings.append(' ') # Blank line

        # Add the top-axis tick labels
        xtick1=xmin
        xtick2=xmin+.25*(xmax-xmin)
        xtick3=xmin+.5*(xmax-xmin)
        xtick4=xmin+.75*(xmax-xmin)
        xtick5=xmax
        plot_strings.append(f"{xtick1:.2e}.......{xtick2:.2e}.......{xtick3:.2e}.......{xtick4:.2e}.......{xtick5:.2e}")

        # Add most of the graph
        for i in range(lpp,-1,-1):
            ytick=None
            if i==lpp or i==round(lpp*3/4) or i==round(lpp/2) or i==round(lpp/4) or i==0:
                ytick=ymin+(i*(ymax-ymin)/lpp)
            prnt_str=''
            for j in range(cpl+1):
                prnt_str += line[i][j]
            if ytick is not None:
                # Add the ytick label to the end of the line
                prnt_str += f"  {ytick:.2f}"
            plot_strings.append(prnt_str)

        # Add the bottom-axis tick labels
        plot_strings.append(f"{xtick1:.2e}.......{xtick2:.2e}.......{xtick3:.2e}.......{xtick4:.2e}.......{xtick5:.2e}")
        
        return plot_strings
    
    def __str__(self):
        """
        Return the ASCII plot as a string.
        :return: string representation of the ASCII plot
        """
        plot_lines = self._qplot()
        return '\n'.join(plot_lines)+'\n'


def qplot(ncur=1,npts=10,x=None,y=None,symbol=None,titl1='',titl2=''):
    """
    Print a simple ASCII plot to stdout.
    :parameter ncur: Number of curves to plot, int
    :parmaeter npts: Number of points per curve, int
    :parameter x: X-coordinates of points, list of ncur arrays of npts floats, [array('f') ... array('f')]
    :parameter y: Y-coordinates of points, list of ncur arrays of npts floats, [array('f') ... array('f')]
    :parameter symbol: Plotting symbol for each curve, array of ncur characters, array('u')
    :parameter titl1: Title line 1, string
    :parameter titl2: Title line 2, string
    :return: None
    """
    plot=TextPlot(ncur,npts,x,y,symbol,titl1,titl2)
    print(plot)
    return None



"""
This module contains unit tests for:
    (1) RheoNetSim and ElongateNetSim classes
    (2) Their associated output classes, RheoNetSimOut and ElongateNetSimOut
    (3) The move_strand_fene_elongational and move_strand_fens_elongational functions
"""


# Standard imports
import unittest
from random import seed
from array import array

# Local imports
from RheologyNetworkModelSimulator.rheonetsim import ElongateNetSim, RheoNetSimOut, ElongateNetSimOut, move_strand_fene_elongational, RheoNetSim, move_strand_fens_elongational
from RheologyNetworkModelSimulator.strand import FENEStrand, FENSStrand
from RheologyNetworkModelSimulator.ensemble import Ensemble
from RheologyNetworkModelSimulator.qplot import TextPlot


class Test_RheoNetSimOut(unittest.TestCase):
    def test_str(self):
        sim_out = RheoNetSimOut()
        sim_out.n_vals.append(10000)
        sim_out.Q_vals.append(0.87)
        sim_out.piYX_vals.append(10.0)
        exp_val = f"{sim_out.n_vals[0]}, {sim_out.piYX_vals[0]}, {sim_out.Q_vals[0]}"
        act_val = str(sim_out)
        self.assertEqual(exp_val, act_val)

    def test_time_step_output(self):
        sim_out = RheoNetSimOut()
        sim_out.n_vals.append(10000)
        sim_out.Q_vals.append(0.87)
        sim_out.piYX_vals.append(10.0)
        exp_val = f"n: {sim_out.n_vals[0]}, piYX: {sim_out.piYX_vals[0]}, Q: {sim_out.Q_vals[0]}"
        act_val = sim_out.time_step_output()
        self.assertEqual(exp_val, act_val)

    def test_final_output(self):
        sim_out = RheoNetSimOut()
        sim_out.n_ave = 0.95
        sim_out.Q_ave = 0.87
        sim_out.piYX_ave = 10.0
        exp_val = f"Avg n: {sim_out.n_ave}\nAvg piYX: {sim_out.piYX_ave}\nAvg Q: {sim_out.Q_ave}\n"
        act_val = sim_out.final_output()
        self.assertEqual(exp_val, act_val)

class Test_ElongateNetSimOut(unittest.TestCase):
    def test_str(self):
        sim_out = ElongateNetSimOut()
        sim_out.n_vals.append(10000)
        sim_out.Q_vals.append(0.87)
        sim_out.piYX_vals.append(10.0)
        sim_out.trout1_vals.append(8.5)
        sim_out.trout2_vals.append(0.2)
        sim_out.trout1_plt=None
        exp_val = f"{sim_out.n_vals[0]}, {sim_out.piYX_vals[0]}, {sim_out.Q_vals[0]}, {sim_out.trout1_vals[0]}, {sim_out.trout2_vals[0]}"
        act_val = str(sim_out)
        self.assertEqual(exp_val, act_val)

    def test_time_step_output(self):
        sim_out = ElongateNetSimOut()
        sim_out.n_vals.append(10000)
        sim_out.Q_vals.append(0.87)
        sim_out.piYX_vals.append(10.0)
        sim_out.trout1_vals.append(8.5)
        sim_out.trout2_vals.append(0.2)
        sim_out.trout1_plt=None
        exp_val = f"n: {sim_out.n_vals[0]}, piYX: {sim_out.piYX_vals[0]}, Q: {sim_out.Q_vals[0]}, Trout 1: {sim_out.trout1_vals[0]}, Trout 2: {sim_out.trout2_vals[0]}"
        act_val = sim_out.time_step_output()
        self.assertEqual(exp_val, act_val)

    def test_final_output(self):
        sim_out = ElongateNetSimOut()
        sim_out.trout1_vals.append(8.5)
        sim_out.trout1_vals.append(7.0)
        sim_out.time_vals.append(0.2)
        sim_out.time_vals.append(0.3)
        sim_out.n_ave = 0.95
        sim_out.Q_ave = 0.87
        sim_out.piYX_ave = 10.0
        sim_out.trout1_ave = 8.5
        sim_out.trout2_ave = 0.2

        # Make a TextPlot object for plotting elongational viscosity vs time
        _symbol = array('u')
        _symbol.append('O')
        _npts=len(sim_out.trout1_vals)
        _x=[sim_out.time_vals]
        _y=[sim_out.trout1_vals]
        _titl1='FENE Network Start-up of Elongational Flow Simulation'
        _titl2='O: Elongational Viscosity vs Time'
        sim_out.trout1_plt = TextPlot(ncur=1, npts=_npts, x=_x, y=_y, symbol=_symbol, titl1=_titl1, titl2=_titl2)

        sim_out.trout1_plt=None
        exp_val = f"Avg n: {sim_out.n_ave}\nAvg piYX: {sim_out.piYX_ave}\nAvg Q: {sim_out.Q_ave}\nAvg Trout 1: {sim_out.trout1_ave}\nAvg Trout 2: {sim_out.trout2_ave}\n\n{sim_out.trout1_plt}"
        act_val = sim_out.final_output()
        self.assertEqual(exp_val, act_val)


class Test_RheoNetSim(unittest.TestCase):
    def test_init(self):
        # strand_mover just needs to be a callable, for this test, which is really testing that
        # _createSimOutObj raises NotImplementedError
        self.assertRaises(NotImplementedError, RheoNetSim, strand_mover=str)


class Test_ElongateNetSim(unittest.TestCase):
    def test_move_fene_strand_elongational(self):
        # def move_strand_fene_elongational(s, eps, gamdot):
        exp_val = (0.09570000000000001, 0.19140000000000001, 0.3258)
        s = FENEStrand(qx=0.1, qy=0.2, qz=0.3, b=100.0, n=2.0, mm=1.0)
        move_strand_fene_elongational(s, 0.001, 100.0)
        act_val = (s.qx, s.qy, s.qz)
        self.assertTupleEqual(exp_val, act_val)

    def test_move_fens_strand_elongational(self):
        # def move_strand_fene_elongational(s, eps, gamdot):
        exp_val = (0.09570000000000001, 0.19140000000000001, 0.3258)
        s = FENEStrand(qx=0.1, qy=0.2, qz=0.3)
        move_strand_fens_elongational(s, 0.001, 100.0)
        act_val = (s.qx, s.qy, s.qz)
        self.assertTupleEqual(exp_val, act_val)

    def test_init(self):
        s = FENEStrand()
        sim = ElongateNetSim({}, s, move_strand_fene_elongational)
        self.assertIsInstance(sim._sim_output, ElongateNetSimOut)

    def test_getOutFileHeader(self):
        s = FENEStrand()
        sim = ElongateNetSim({}, s, move_strand_fene_elongational)
        exp_val = "Time, Strands, <Q>, piYX, trouton_1, trouton_2"
        act_val = sim._getOutFileHeader()
        self.assertEqual(exp_val, act_val)

    def test_computeTimestepOutputs_before_ss(self):
        fene_b = 100.0 # All results in J. Chem. Phys article
        fene_n = 2.0 # Don't change
        fene_mm = 1.0 # Don't change
        sim_input = {}
        sim_input['gamdot'] = 10.0
        sim_input['begstrand'] = 100
        sim_input['eps'] = 0.001
        sim_input['steps'] = int(0.01 / sim_input['eps'])
        sim_input['outfile'] = 'testfenetrout.out'
        s1 = FENEStrand(qx=0.1, qy=0.2, qz=0.3, b=fene_b, n=fene_n, mm=fene_mm)
        s2 = FENEStrand(qx=0.3, qy=0.1, qz=0.2, b=fene_b, n=fene_n, mm=fene_mm)
        sim = ElongateNetSim(sim_input, s1, move_strand_fene_elongational)
        we = Ensemble([s1, s2])
        exp_val = ElongateNetSimOut()
        exp_val.n_vals.append(2)
        exp_val.piYX_vals.append(0.058139535)
        exp_val.Q_vals.append(0.374165739)
        exp_val.time_vals.append(0.007)
        exp_val.trout1_vals.append(0.003488372)
        exp_val.trout2_vals.append(-0.005813953)
        exp_val.n_ave = 0
        exp_val.piYX_ave = 0
        exp_val.Q_ave = 0
        exp_val.trout1_ave = 0
        exp_val.trout2_ave = 0
        sim._computeTimestepOutputs(we, 0.007)
        act_val = sim._sim_output
        self.assertAlmostEqual(exp_val.n_vals[0], act_val.n_vals[0], 9)
        self.assertAlmostEqual(exp_val.Q_vals[0], act_val.Q_vals[0], 9)
        self.assertAlmostEqual(exp_val.time_vals[0], act_val.time_vals[0], 9)
        self.assertAlmostEqual(exp_val.piYX_vals[0], act_val.piYX_vals[0], 9)
        self.assertAlmostEqual(exp_val.trout1_vals[0], act_val.trout1_vals[0], 9)
        self.assertAlmostEqual(exp_val.trout2_vals[0], act_val.trout2_vals[0], 9)
        self.assertAlmostEqual(exp_val.n_ave, act_val.n_ave, 9)
        self.assertAlmostEqual(exp_val.piYX_ave, act_val.piYX_ave, 9)
        self.assertAlmostEqual(exp_val.Q_ave, act_val.Q_ave, 9)
        self.assertAlmostEqual(exp_val.trout1_ave, act_val.trout1_ave, 9)
        self.assertAlmostEqual(exp_val.trout2_ave, act_val.trout2_ave, 9)

    def test_computeTimestepOutputs_after_ss(self):
        fene_b = 100.0 # All results in J. Chem. Phys article
        fene_n = 2.0 # Don't change
        fene_mm = 1.0 # Don't change
        sim_input = {}
        sim_input['gamdot'] = 10.0
        sim_input['begstrand'] = 100
        sim_input['eps'] = 0.001
        sim_input['steps'] = int(0.01 / sim_input['eps'])
        sim_input['outfile'] = 'testfenetrout.out'
        s1 = FENEStrand(qx=0.1, qy=0.2, qz=0.3, b=fene_b, n=fene_n, mm=fene_mm)
        s2 = FENEStrand(qx=0.3, qy=0.1, qz=0.2, b=fene_b, n=fene_n, mm=fene_mm)
        sim = ElongateNetSim(sim_input, s1, move_strand_fene_elongational)
        we = Ensemble([s1, s2])
        exp_val = ElongateNetSimOut()
        exp_val.n_vals.append(2)
        exp_val.piYX_vals.append(0.058139535)
        exp_val.Q_vals.append(0.374165739)
        exp_val.time_vals.append(0.008)
        exp_val.trout1_vals.append(0.003488372)
        exp_val.trout2_vals.append(-0.005813953)
        exp_val.n_ave = 2/sim_input['begstrand']
        exp_val.piYX_ave = 0.058139535
        exp_val.Q_ave = 0.374165739
        exp_val.trout1_ave = 0.003488372
        exp_val.trout2_ave = -0.005813953
        sim._computeTimestepOutputs(we, 0.008)
        act_val = sim._sim_output
        self.assertAlmostEqual(exp_val.n_vals[0], act_val.n_vals[0], 9)
        self.assertAlmostEqual(exp_val.Q_vals[0], act_val.Q_vals[0], 9)
        self.assertAlmostEqual(exp_val.time_vals[0], act_val.time_vals[0], 9)
        self.assertAlmostEqual(exp_val.piYX_vals[0], act_val.piYX_vals[0], 9)
        self.assertAlmostEqual(exp_val.n_ave, act_val.n_ave, 9)
        self.assertAlmostEqual(exp_val.piYX_ave, act_val.piYX_ave, 9)
        self.assertAlmostEqual(exp_val.Q_ave, act_val.Q_ave, 9)
        self.assertAlmostEqual(exp_val.trout1_ave, act_val.trout1_ave, 9)
        self.assertAlmostEqual(exp_val.trout2_ave, act_val.trout2_ave, 9)

    def test_computeTimestepOutputs_after_ss_finalize(self):
        fene_b = 100.0 # All results in J. Chem. Phys article
        fene_n = 2.0 # Don't change
        fene_mm = 1.0 # Don't change
        sim_input = {}
        sim_input['gamdot'] = 10.0
        sim_input['begstrand'] = 100
        sim_input['eps'] = 0.001
        sim_input['steps'] = int(0.01 / sim_input['eps'])
        sim_input['outfile'] = 'testfenetrout.out'
        s1 = FENEStrand(qx=0.1, qy=0.2, qz=0.3, b=fene_b, n=fene_n, mm=fene_mm)
        s2 = FENEStrand(qx=0.3, qy=0.1, qz=0.2, b=fene_b, n=fene_n, mm=fene_mm)
        sim = ElongateNetSim(sim_input, s1, move_strand_fene_elongational)
        we = Ensemble([s1, s2])
        exp_val = ElongateNetSimOut()
        exp_val.n_vals.append(2)
        exp_val.piYX_vals.append(0.058139535)
        exp_val.Q_vals.append(0.374165739)
        exp_val.time_vals.append(0.008)
        exp_val.trout1_vals.append(0.003488372)
        exp_val.trout2_vals.append(-0.005813953)
        exp_val.n_ave = 2/sim_input['begstrand']
        exp_val.piYX_ave = 0.058139535
        exp_val.Q_ave = 0.374165739
        exp_val.trout1_ave = 0.003488372
        exp_val.trout2_ave = -0.005813953
        # Manipulate internal state so that data is present when finalize branch is taken
        sim._sim_output.n_vals.append(2)
        sim._sim_output.piYX_vals.append(0.058139535)
        sim._sim_output.Q_vals.append(0.374165739)
        sim._sim_output.time_vals.append(0.008)
        sim._sim_output.trout1_vals.append(0.003488372)
        sim._sim_output.trout2_vals.append(-0.005813953)
        sim._sim_output.n_ave = exp_val.n_ave
        sim._sim_output.piYX_ave = exp_val.piYX_ave
        sim._sim_output.Q_ave = exp_val.Q_ave
        sim._sim_output.trout1_ave = exp_val.trout1_ave
        sim._sim_output.trout2_ave = exp_val.trout2_ave
        # Now modify exp_val to reflect finalize calculations
        exp_val.n_ave = exp_val.n_ave/round(0.25*int(0.01 / sim_input['eps']))
        exp_val.piYX_ave = exp_val.piYX_ave/round(0.25*int(0.01 / sim_input['eps']))
        exp_val.Q_ave = exp_val.Q_ave/round(0.25*int(0.01 / sim_input['eps']))
        exp_val.trout1_ave = exp_val.trout1_ave/round(0.25*int(0.01 / sim_input['eps']))
        exp_val.trout2_ave = exp_val.trout2_ave/round(0.25*int(0.01 / sim_input['eps']))
        # Now call with finalize=True
        sim._computeTimestepOutputs(we, 0.008, True)
        act_val = sim._sim_output
        self.assertAlmostEqual(exp_val.n_vals[0], act_val.n_vals[0], 9)
        self.assertAlmostEqual(exp_val.Q_vals[0], act_val.Q_vals[0], 9)
        self.assertAlmostEqual(exp_val.time_vals[0], act_val.time_vals[0], 9)
        self.assertAlmostEqual(exp_val.piYX_vals[0], act_val.piYX_vals[0], 9)
        self.assertAlmostEqual(exp_val.n_ave, act_val.n_ave, 9)
        self.assertAlmostEqual(exp_val.piYX_ave, act_val.piYX_ave, 9)
        self.assertAlmostEqual(exp_val.Q_ave, act_val.Q_ave, 9)
        self.assertAlmostEqual(exp_val.trout1_ave, act_val.trout1_ave, 9)
        self.assertAlmostEqual(exp_val.trout2_ave, act_val.trout2_ave, 9)

    def test_run_sim_elongate_fene(self):
        seed(1234567890)

        exp_val = {'Qave': 0.16110272799636224, 'n/no': 0.99, 'piYX': 0.07245743211970784, 'trouton 1': 0.04737104142315748, 'trouton 2': 0.0021919416527666088}

        fene_b = 100.0 # All results in J. Chem. Phys article
        fene_n = 2.0 # Don't change
        fene_mm = 1.0 # Don't change

        sim_input = {}
        sim_input['gamdot'] = 10.0
        sim_input['begstrand'] = 100
        sim_input['eps'] = 0.001
        sim_input['steps'] = int(0.01 / sim_input['eps'])
        sim_input['outfile'] = 'testfenetrout.out'

        _proto_strand = FENEStrand(qx=0, qy=0, qz=0, b=fene_b, n=fene_n, mm=fene_mm)
        sim = ElongateNetSim(sim_input, _proto_strand, move_strand_fene_elongational)
        act_val = sim.run_sim()
        print(act_val)
        self.assertAlmostEqual(exp_val['Qave'], act_val.Q_ave, 15)
        self.assertAlmostEqual(exp_val['n/no'], act_val.n_ave, 15)
        self.assertAlmostEqual(exp_val['piYX'], act_val.piYX_ave,15)
        self.assertAlmostEqual(exp_val['trouton 1'], act_val.trout1_ave, 15)
        self.assertAlmostEqual(exp_val['trouton 2'], act_val.trout2_ave, 15)

    def test_run_sim_elongate_fens(self):
        seed(1234567890)

        exp_val = {'Qave': 1.5689606851693894, 'n/no': 1.01, 'piYX': -0.0614371945729071, 'trouton 1': -0.0009294645015770726, 'trouton 2': -0.018659087231995483}

        sim_input = {}
        sim_input['gamdot'] = 10.0
        sim_input['begstrand'] = 100
        sim_input['eps'] = 0.001
        sim_input['steps'] = int(0.01 / sim_input['eps'])
        sim_input['outfile'] = 'testfenstrout.out'

        _proto_strand = FENSStrand(qx=0, qy=0, qz=0)
        sim = ElongateNetSim(sim_input, _proto_strand, move_strand_fens_elongational)
        act_val = sim.run_sim()
        print(act_val)
        self.assertAlmostEqual(exp_val['Qave'], act_val.Q_ave, 15)
        self.assertAlmostEqual(exp_val['n/no'], act_val.n_ave, 15)
        self.assertAlmostEqual(exp_val['piYX'], act_val.piYX_ave,15)
        self.assertAlmostEqual(exp_val['trouton 1'], act_val.trout1_ave, 15)
        self.assertAlmostEqual(exp_val['trouton 2'], act_val.trout2_ave, 15)


if __name__ == '__main__':
    unittest.main()



"""
This module's __main__ first asks the user how they want to use the Network Model Simulator, then
collects simulation input interactively from the user, executes a simulation, and prints some results.

Exported classes:
    None

Exported functions:
    __main__: Requests simulation type, collects input, runs simulation, prints results output
    Fene_Trout: Run an elongatonal flow simulation using the FENE network model.
    Fens_Trout: Run an elongatonal flow simulation using the FENS network model.
    debug: Run a debugging scenario (currently does nothing).

Exported exceptions:
    None
"""


# local imports
from UserResponseCollector.UserQueryCommand import askForInt, askForFloat, askForPathSave, askForPathSave, askForStr, askForMenuSelection
from RheologyNetworkModelSimulator.rheonetsim import ElongateNetSim, move_strand_fene_elongational, move_strand_fens_elongational
from RheologyNetworkModelSimulator.strand import FENEStrand, FENSStrand


def debug():
    """
    Run a debugging scenario.
    """
    return None


def Fene_Trout():
    """
    Run an elongatonal flow simulation using the FENE network model.
    """
    gamdot = askForFloat('Enter the elongation rate (suggest 1.0)', minimum=0.0)
    begstrand = askForInt('Enter the number of strands in the equilibrium ensemble (suggest 10000)', minimum=1)
    eps = askForFloat('Enter the time step size (suggest 0.001)', minimum=1e-5, maximum=0.1)
    endt = askForFloat('Enter the end time of the simulation (suggest 5.0). Steady-state results will be averaged over the last quarter of this time.', minimum=10*eps)
    steps = int(endt / eps)
    # Suggested value for b matches all results in J. Chem. Phys. article
    fene_b = askForFloat('Enter the FENE Parameter b (suggest 100.0)', minimum=1.0)
    outfil = askForPathSave('Path of output file')
    banner = askForStr('Type an identifier for ouput')

    # Package the required simulation input into a dictionary
    sim_input = {}
    sim_input['gamdot'] = gamdot
    sim_input['begstrand'] = begstrand
    sim_input['eps'] = eps
    sim_input['steps'] = steps
    sim_input['outfile'] = outfil
    # TODO: Investigate if this is actually used
    sim_input['banner'] = ''

    # Initialize the simulation
    _proto_strand = FENEStrand(qx=0, qy=0, qz=0, b=fene_b, n=2.0, mm=1.0) # Don't change n or mm
    sim = ElongateNetSim(sim_input, _proto_strand, move_strand_fene_elongational)

    # Execute the simulation
    so=sim.run_sim()

    # Print some output from the simulation
    print(f"\n**Final Simulation Results**\n{so.final_output()}")

    return None


def Fens_Trout():
    """
    Run an elongatonal flow simulation using the FENS network model.
    """
    gamdot = askForFloat('Enter the elongation rate (suggest 1.0)', minimum=0.0)
    begstrand = askForInt('Enter the number of strands in the equilibrium ensemble (suggest 10000)', minimum=1)
    eps = askForFloat('Enter the time step size (suggest 0.001)', minimum=1e-5, maximum=0.1)
    endt = askForFloat('Enter the end time of the simulation (suggest 5.0). Steady-state results will be averaged over the last quarter of this time.', minimum=10*eps)
    steps = int(endt / eps)
    outfil = askForPathSave('Path of output file')
    banner = askForStr('Type an identifier for ouput')

    # Package the required simulation input into a dictionary
    sim_input = {}
    sim_input['gamdot'] = gamdot
    sim_input['begstrand'] = begstrand
    sim_input['eps'] = eps
    sim_input['steps'] = steps
    sim_input['outfile'] = outfil
    # TODO: Investigate if this is actually used
    sim_input['banner'] = ''

    # Initialize the simulation
    _proto_strand = FENSStrand(qx=0, qy=0, qz=0)
    sim = ElongateNetSim(sim_input, _proto_strand, move_strand_fens_elongational)

    # Execute the simulation
    so=sim.run_sim()

    # Print some output from the simulation
    print(f"\n**Final Simulation Results**\n{so.final_output()}")

    return None


if __name__ == '__main__':
    """
    Query the user for how they wish to use the Network Model Simulator, and then launch that usage.
    This includes a "debug" usage to set up what ever situation is needed for debugging, since I can't seem to reliably debug unit tests.
    """
    print('---------------------------------------------')
    print('-----     Network Model Simulator       -----')
    print('---------------------------------------------')
    
    # Build a query for the user to obtain their choice of how to user the simulator
    query_preface = 'How do you want to use the simulator?'
    query_dic = {'q':'Quit', '1':'FENE Elongational', '2':'FENS Elongational', 'd':'Debug Scenario'}
    response = askForMenuSelection(query_preface, query_dic)
    
    while response != 'q':
        
        match response:
            
            case '1':
                Fene_Trout()
                
            case '2':
                Fens_Trout()
                
            case 'd':
                debug()
        
        print('--------------------')
        response = askForMenuSelection(query_preface, query_dic)



"""
This module defines the class Ensemble, which represents an ensemble (collection) of network strands in a network model for polymer melt rheology.

Exported classes:
    Ensemble: Class to represent an ensemble of network strands in a polymer network.

Exported functions:


Exported exceptions:
    None
"""


# Standard imports
from math import sqrt


# Local imports
from RheologyNetworkModelSimulator.strand import Strand


class Ensemble(object):
    """
    This class represents an ensemble of network strands in a polymer network. This class is iterable over its Strand objects.
    """
    def __init__(self, strands):
        """
        Construct an Ensemble with a given list of Strand objects.
        :param strands: List of Strand objects representing the ensemble.
        """
        assert(isinstance(strands, list))
        assert(isinstance(strands[0], Strand))  # Test the first element to (roughly) confirm list of Strand objects.
        self._strands = list(strands)

    @property
    def strands(self):
        """
        Get a copy of the list of Strand objects in the ensemble.
        :return: A copy of the list of Strand objects in the ensemble.
        """
        return list(self._strands)
        
    def write_ensemble_to_file(self, filename):
        """
        Write a CSV formatted text file with the internal coordinates and length of each strand  in list.
        :param filename: Name of file to write to, as string.
        :return: None.
        """
        f = open(filename, 'w')
        f.write('qx, qy, qz, length\n')
        i = 1
        for s in self._strands:
            f.write('%i, %f, %f, %f, %f\n' % (i, s.qx, s.qy, s.qz, sqrt(s.str_len_sqr())))
            i = i +1
        f.close()
        return None

    def q_ave(self):
        """
        Compute the average non-dimensional length of a strand in the network ensemble.
        :return: Average non-dimensional length of a strand in the network ensemble, as float
        Strand length is non-dimensionalized consistent self._strands.
        """
        total_q = 0.0
        for s in self._strands:
            total_q = total_q + sqrt(s.str_len_sqr())
        return total_q / len(self._strands)

    def ensemble_stress(self):
        """
        Compute the components of the total stress tensor for the network,
        by summing over the strands in the network ensemble.  Stress tensor is divided by NokT.
        :param e: List of strands (the "ensemble"), as [Strand objects].
        :return: Tuple (XX, YY, ZZ, YX) of components of total stress tensor of the network, divided by NokT, as (float, float, float, float)
        """
        piXX = 0.0
        piYY = 0.0
        piZZ = 0.0
        piYX = 0.0
        for s in self._strands:
            piXX = piXX + s.stress_XX()
            piYY = piYY + s.stress_YY()
            piZZ = piZZ + s.stress_ZZ()
            piYX = piYX + s.stress_YX()
        return (piXX, piYY, piZZ, piYX)

    # Implement methods so that an Ensemble can be treated as a container type. See Section 3.3.7. Emulating container types in
    # https://docs.python.org/3/reference/datamodel.html#classgetitem-versus-getitem
    
    def __iter__(self):
        return iter(self._strands)
    
    def __len__(self):
        return len(self._strands)
    
    def __getitem__(self, item):
        if item >= len(self):
            raise IndexError("Ensemble index out of range")
        return self._strands[item]

    def remove(self, value):
        for i in self._strands:
            if i == value:
                self._strands.remove(i)
                break
        return

    def append(self, value):
        self._strands.append(value)
        return
